import os
import sys
from shipyard_bigquery import BigQueryClient


def main():
    google_client = BigQueryClient(os.environ.get("GOOGLE_APPLICATION_CREDENTIALS"))
    try:
        con = google_client.connect()
        google_client.logger.info("Successfully established a connection")
        sys.exit(0)
    except Exception as e:
        google_client.logger.error("Could not establish a connection")
        sys.exit(1)


if __name__ == "__main__":
    main()



from typing import Dict, Union, List
from shipyard_templates import ShipyardLogger


logger = ShipyardLogger.get_logger()


def validate_data_types(
    data_types: Union[List[List[str]], List[Dict[str, str]]]
) -> bool:
    BIGQUERY_DATA_TYPES = {
        "STRING",
        "BYTES",
        "INT64",
        "FLOAT64",
        "BOOL",
        "TIMESTAMP",
        "DATE",
        "TIME",
        "DATETIME",
        "GEOGRAPHY",
        "JSON",
        "INTERVAL",
        "NUMERIC",
        "BIGNUMERIC",
        "STRUCT",
        "TIME",
        ##NOTE: the following below should be supported too
        "INT",
        "SMALLINT",
        "INTEGER",
        "BIGINT",
        "TINYINT",
        "BYTEINT",
    }
    """ Helper function to validate whether the provided data types are real BigQuery data types

    Args:
        data_types: The inputted data types to be used in the upload

    Returns: True if valid, False otherwise
        
    """
    for item in data_types:
        if isinstance(item, list):
            if str(item[1]).upper() not in BIGQUERY_DATA_TYPES:
                logger.debug(f"Type {item[1]} is not a valid BigQuery data type")
                return False
        elif isinstance(item, dict):
            if str(item["type"]).upper() not in BIGQUERY_DATA_TYPES:
                logger.debug(f"Type {item['type']} is not a valid BigQuery data type")
                return False
    else:
        return True



from shipyard_templates import ExitCodeException, GoogleDatabase

EXIT_CODE_FETCH_ERROR = 101
EXIT_CODE_QUERY_ERROR = 102
EXIT_CODE_DOWNLOAD_TO_GCS_ERROR = 103
EXIT_CODE_TEMP_TABLE_CREATION_ERROR = 104
EXIT_CODE_SCHEMA_FORMATTING_ERROR = 105
EXIT_CODE_SCHEMA_VALIDATION_ERROR = 106


class InvalidSchema(ExitCodeException):
    def __init__(self, message: str):
        self.message = message
        self.exit_code = GoogleDatabase.EXIT_CODE_INVALID_SCHEMA


class FetchError(ExitCodeException):
    def __init__(self, message: str):
        self.message = message
        self.exit_code = EXIT_CODE_FETCH_ERROR


class QueryError(ExitCodeException):
    def __init__(self, message: str):
        self.message = message
        self.exit_code = EXIT_CODE_QUERY_ERROR


class DownloadToGcsError(ExitCodeException):
    def __init__(self, message: str):
        self.message = message
        self.exit_code = EXIT_CODE_DOWNLOAD_TO_GCS_ERROR


class TempTableCreationError(ExitCodeException):
    def __init__(self, message: str):
        self.message = message
        self.exit_code = EXIT_CODE_TEMP_TABLE_CREATION_ERROR


class SchemaFormatError(ExitCodeException):
    def __init__(self, message: str):
        self.message = message
        self.exit_code = EXIT_CODE_SCHEMA_FORMATTING_ERROR


class SchemaValidationError(ExitCodeException):
    def __init__(self, message: str):
        self.message = message
        self.exit_code = EXIT_CODE_SCHEMA_VALIDATION_ERROR



import os
import re
import argparse
import ast
import sys

from shipyard_templates import ShipyardLogger, ExitCodeException
from shipyard_bigquery import BigQueryClient
from shipyard_bigquery.utils.exceptions import InvalidSchema, SchemaFormatError
import shipyard_bp_utils as shipyard

logger = ShipyardLogger.get_logger()


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--dataset", dest="dataset", required=True)
    parser.add_argument("--table", dest="table", required=True)
    parser.add_argument("--service-account", dest="service_account", required=True)
    parser.add_argument(
        "--upload-type",
        dest="upload_type",
        default="append",
        choices={"append", "overwrite"},
        required=False,
    )
    parser.add_argument(
        "--source-file-name-match-type",
        dest="source_file_name_match_type",
        default="exact_match",
        choices={"exact_match", "regex_match"},
        required=False,
    )
    parser.add_argument("--source-file-name", dest="source_file_name", required=True)
    parser.add_argument(
        "--source-folder-name", dest="source_folder_name", default="", required=False
    )
    parser.add_argument("--schema", dest="schema", default="", required=False)
    parser.add_argument(
        "--skip-header-rows", dest="skip_header_rows", default="", required=False
    )
    parser.add_argument(
        "--quoted-newline", dest="quoted_newline", default="FALSE", required=False
    )
    args = parser.parse_args()
    return args


def main():
    try:
        args = get_args()
        dataset = args.dataset
        table = args.table
        upload_type = args.upload_type
        file_name = args.source_file_name
        folder_name = args.source_folder_name
        full_path = shipyard.files.combine_folder_and_file_name(
            folder_name=f"{os.getcwd()}/{folder_name}", file_name=file_name
        )
        match_type = args.source_file_name_match_type
        schema = None if args.schema == "" else ast.literal_eval(args.schema)
        quoted_newline = args.quoted_newline.strip().upper() == "TRUE"

        skip_header_rows = (
            None if args.skip_header_rows == "" else args.skip_header_rows
        )

        if skip_header_rows:
            skip_header_rows = int(args.skip_header_rows)

        if schema and not skip_header_rows:
            logger.warning(
                f"Skip Header Rows was not provided but a schema was defined. Setting the Skip Header Rows to 1 to properly load"
            )
            skip_header_rows = 1

        client = BigQueryClient(args.service_account)
        client.connect()
        logger.info("Successfully connected to BigQuery")

        if match_type == "regex_match":
            file_names = shipyard.files.find_all_local_file_names(folder_name)
            matching_file_names = shipyard.files.find_all_file_matches(
                file_names, re.compile(file_name)
            )
            logger.info(
                f"{len(matching_file_names)} files found. Preparing to upload..."
            )

            for index, file_name in enumerate(matching_file_names, start=1):
                logger.info(f"Uploading file {index} of {len(matching_file_names)}")
                client.upload(
                    file=file_name,
                    dataset=dataset,
                    table=table,
                    upload_type=upload_type,
                    skip_header_rows=skip_header_rows,
                    schema=schema,
                    quoted_newline=quoted_newline,
                )
                upload_type = "append"
        else:
            client.upload(
                file=full_path,
                dataset=dataset,
                table=table,
                upload_type=upload_type,
                skip_header_rows=skip_header_rows,
                schema=schema,
                quoted_newline=quoted_newline,
            )
            logger.info(f"Successfully loaded {full_path} to {dataset}.{table}")
    except FileNotFoundError as fe:
        logger.error(str(fe))
        sys.exit(BigQueryClient.EXIT_CODE_FILE_NOT_FOUND)
    except (InvalidSchema, SchemaFormatError, ExitCodeException) as ec:
        logger.error(ec.message)
        sys.exit(ec.exit_code)
    except Exception as e:
        logger.error(f"Error in uploading file to BigQuery: {str(e)}")
        sys.exit(BigQueryClient.EXIT_CODE_INVALID_UPLOAD_VALUE)


if __name__ == "__main__":
    main()



import argparse
import sys
import os

from shipyard_bigquery import BigQueryClient
from shipyard_bigquery.utils.exceptions import DownloadToGcsError
from shipyard_templates import ShipyardLogger, ExitCodeException
import shipyard_bp_utils as shipyard

logger = ShipyardLogger.get_logger()


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--query", dest="query", required=True)
    parser.add_argument("--service-account", dest="service_account", required=True)
    parser.add_argument("--bucket-name", dest="bucket_name", required=True)
    parser.add_argument(
        "--destination-file-name",
        dest="destination_file_name",
        default="output.csv",
        required=True,
    )
    parser.add_argument(
        "--destination-folder-name",
        dest="destination_folder_name",
        default="",
        required=False,
    )
    args = parser.parse_args()
    return args


def main():
    try:
        args = get_args()
        destination_file_name = args.destination_file_name
        target_folder = args.destination_folder_name or os.getcwd()
        target_path = shipyard.files.combine_folder_and_file_name(
            folder_name=target_folder, file_name=destination_file_name
        )
        client = BigQueryClient(args.service_account)
        client.connect()
        logger.info("Successfully connected to BigQuery")
        logger.debug(f"Service account email is {client.email}")

        logger.info("Beginning job to store query results in GCS")
        client.download_to_gcs(
            query=args.query, bucket_name=args.bucket_name, path=target_path
        )
    except (ExitCodeException, DownloadToGcsError) as ec:
        logger.error(ec.message)
        sys.exit(ec.exit_code)
    except Exception as e:
        logger.error(f"An error occurred in downloading query results to GCS: {str(e)}")
        sys.exit(BigQueryClient.EXIT_CODE_NO_RESULTS)
    else:
        logger.info("Successfully stored query results in GCS")


if __name__ == "__main__":
    main()



import json
from google.cloud.bigquery.table import RowIterator
import pandas as pd
from typing import Optional, Dict, Union, List
from google.cloud import bigquery
from google.oauth2 import service_account
from shipyard_templates import GoogleDatabase, ShipyardLogger, ExitCodeException
from shipyard_bigquery.utils.exceptions import (
    DownloadToGcsError,
    FetchError,
    InvalidSchema,
    QueryError,
    SchemaFormatError,
    SchemaValidationError,
    TempTableCreationError,
)
from shipyard_bigquery.utils import utils

logger = ShipyardLogger.get_logger()


class BigQueryClient(GoogleDatabase):
    def __init__(self, service_account: str) -> None:
        self.service_account = service_account
        super().__init__(service_account)

    @property
    def json_creds(self):
        return json.loads(self.service_account)

    @property
    def credentials(self):
        return service_account.Credentials.from_service_account_info(self.json_creds)

    @property
    def email(self):
        return self.credentials.service_account_email

    def connect(self):
        conn = bigquery.Client(credentials=self.credentials)
        self.conn = conn
        return self

    def execute_query(self, query: str) -> RowIterator:
        """Executes a query and returns the results

        Args:
            query: The query to execute

        Returns: The query results as a pandas dataframe

        """
        try:
            query_job = self.conn.query(query)
            results = query_job.result()
        except Exception as e:
            raise QueryError(f"Error in executing query: {str(e)}")
        else:
            return results

    def fetch(self, query: str) -> pd.DataFrame:
        """Returns the results of a query to a pandas dataframe

        Args:
            query: The query to fetch

        Returns: The query results as a pandas dataframe

        """
        try:
            df = self.conn.query(query).to_dataframe()
        except Exception as e:
            raise FetchError(f"Error in fetching query: {str(e)}")
        else:
            return df

    def upload(
        self,
        file: str,
        dataset: str,
        table: str,
        upload_type: str,
        skip_header_rows: Optional[int] = None,
        schema: Optional[Union[List[List], Dict[str, str]]] = None,
        quoted_newline: bool = False,
    ):
        """Upload a file to a table in BigQuery
        Args:
            file: The file to load
            dataset: The name of the dataset in BigQuery to upload to
            table: The name of the table to write to
            upload_type: Whether to append to or replace the data. Choices are 'overwrite' and 'append'
            schema: The optional schema of the table to be loaded
            skip_header_rows: Whether to skip the header row
            quoted_newline: Whether newline characters should be quoted
        """
        try:
            dataset_ref = self.conn.dataset(dataset)
            table_ref = dataset_ref.table(table)
            job_config = bigquery.LoadJobConfig()

            if upload_type == "overwrite":
                job_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE
            else:
                job_config.write_disposition = bigquery.WriteDisposition.WRITE_APPEND
            job_config.source_format = bigquery.SourceFormat.CSV
            job_config.autodetect = True  # infer the schema
            if skip_header_rows:
                job_config.skip_leading_rows = skip_header_rows
            if schema:
                # TODO: add validation check for schema type
                # TODO: during validation, identify which datatype is bad
                if not utils.validate_data_types(schema):
                    raise SchemaValidationError(
                        "Inputted schema contains an invalid data type. Run with LOG_LEVEL set to DEBUG for more information"
                    )
                logger.debug(f"Schema is {schema}")
                job_config.autodetect = False
                job_config.schema = self._format_schema(schema)
            if quoted_newline:
                job_config.allow_quoted_newlines = True
            with open(file, "rb") as source_file:
                job = self.conn.load_table_from_file(
                    source_file, table_ref, job_config=job_config
                )
            job.result()
        except SchemaValidationError:
            raise
        except SchemaFormatError:
            raise
        except InvalidSchema:
            raise
        except Exception as e:
            raise ExitCodeException(
                f"An error occurred when attempting to upload to BigQuery: {str(e)}",
                self.EXIT_CODE_INVALID_UPLOAD_VALUE,
            )

    def download_to_gcs(self, query: str, bucket_name: str, path: Optional[str] = None):
        try:
            project_id, dataset_id, table_id, location = self._create_temp_table(query)
            dataset_ref = bigquery.DatasetReference(
                project=project_id, dataset_id=dataset_id
            )
            table_ref = dataset_ref.table(table_id)
            dest_uri = f"gs://{bucket_name}/{path}"
            self.conn.extract_table(table_ref, dest_uri, location=location).result()
        except TempTableCreationError as te:
            raise DownloadToGcsError(
                f"Error in executing the query and storing in a temp file: {te.message}"
            )
        except Exception as e:
            raise DownloadToGcsError(f"Error downloading file to GCS: {str(e)}")

    def _format_schema(
        self, schema: Union[List[List[str]], Dict[str, List[str]]]
    ) -> List[bigquery.SchemaField]:
        """Helper function to format the schema appropriately for BigQuery

        Args:
            schema: The representation inputted as either a list of lists (for backwards compatibility) or preferably JSON

        Returns: The formatted schema

        """
        formatted_schema = []
        try:
            for item in schema:
                if isinstance(item, list):
                    formatted_schema.append(bigquery.SchemaField(*item))
                elif isinstance(item, dict):
                    formatted_schema.append(bigquery.SchemaField.from_api_repr(item))
                else:
                    raise InvalidSchema(
                        "Format of inputted schema is incorrect, this should preferably be a JSON representation or a List of Lists. For additional information and examples, visit https://cloud.google.com/bigquery/docs/schemas#specifying_a_json_schema_file"
                    )
        except Exception as e:
            raise SchemaFormatError(
                f"Error in preparing the inputted schema to the approrpriate BigQuery format: {str(e)}"
            )
        else:
            logger.debug(f"Formatted schema is {formatted_schema}")
            return formatted_schema

    def _create_temp_table(self, query: str):
        """Helper function to execute a query and store the results in a temporary table

        Args:
            query: The query to execute

        Returns: The metadata of the temp table

        """
        try:
            data = self.conn.query(query)
            data.result()
            temp_table_ids = data._properties["configuration"]["query"][
                "destinationTable"
            ]
            location = data._properties["jobReference"]["location"]
            project_id = temp_table_ids.get("projectId")
            dataset_id = temp_table_ids.get("datasetId")
            table_id = temp_table_ids.get("tableId")
        except Exception as e:
            raise TempTableCreationError(
                f"Error in storing query results in temp table: {str(e)}"
            )
        else:
            return project_id, dataset_id, table_id, location



import os
import argparse
import sys
import shipyard_bp_utils as shipyard


from shipyard_templates import ShipyardLogger, ExitCodeException
from shipyard_bigquery import BigQueryClient


logger = ShipyardLogger.get_logger()


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--query", dest="query", required=True)
    parser.add_argument("--service-account", dest="service_account", required=True)
    parser.add_argument(
        "--destination-file-name",
        dest="destination_file_name",
        default="output.csv",
        required=True,
    )
    parser.add_argument(
        "--destination-folder-name",
        dest="destination_folder_name",
        default="",
        required=False,
    )
    args = parser.parse_args()
    return args


def main():
    try:
        args = get_args()
        target_folder = args.destination_folder_name or os.getcwd()
        shipyard.files.create_folder_if_dne(target_folder)
        target_path = shipyard.files.combine_folder_and_file_name(
            folder_name=target_folder, file_name=args.destination_file_name
        )
        client = BigQueryClient(args.service_account)
        client.connect()
        logger.info("Successfully connected to BigQuery")
        logger.debug(f"Query is {args.query}")
        df = client.fetch(args.query)
        logger.debug(f"Shape of the data is {df.shape}")
        df.to_csv(target_path, index=False)
    except ExitCodeException as ec:
        logger.error(ec.message)
        sys.exit(ec.exit_code)
    except Exception as e:
        logger.error(
            f"Error in fetching query results and writing results to {target_path}: {str(e)}"
        )
        sys.exit(BigQueryClient.EXIT_CODE_NO_RESULTS)

    else:
        logger.info(f"Successfully wrote query results to {target_path}")


if __name__ == "__main__":
    main()



import argparse
import sys
from shipyard_templates import ShipyardLogger, ExitCodeException
from shipyard_bigquery import BigQueryClient
from shipyard_bigquery.utils.exceptions import QueryError, EXIT_CODE_QUERY_ERROR

logger = ShipyardLogger.get_logger()


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--query", dest="query", required=True)
    parser.add_argument("--service-account", dest="service_account", required=True)
    args = parser.parse_args()
    return args


def main():
    try:
        args = get_args()
        query = args.query
        logger.debug(f"Query is {query}")
        client = BigQueryClient(args.service_account)
        client.connect()
        logger.info("Successfully connected to BigQuery")
        logger.debug(f"Service account email is {client.email}")
        client.execute_query(query)
    except (QueryError, ExitCodeException) as ec:
        logger.error(ec.message)
        sys.exit(ec.exit_code)
    except Exception as e:
        logger.error(f"Error in executing query: {str(e)}")
        sys.exit(EXIT_CODE_QUERY_ERROR)
    else:
        logger.info("Successfully executed query")


if __name__ == "__main__":
    main()
